""" Using a nearest-neighbor approach on the CIFAR-10 dataset
    Authors: Alessandro Giusti and Giacomo Boracchi
    Contact: alessandrog@idsia.ch 
    
    Note: this file is meant to be executed cell by cell.
    We suggest using the spyder IDE, which supports cell mode
    You execute a cell using CTRL+Enter """
    
#%%  Read in our data
    
# you can download the data from here: https://www.cs.toronto.edu/~kriz/cifar.html
# In particular, we are using the CIFAR-10 dataset in the Python version:
# https://www.cs.toronto.edu/~kriz/cifar-10-python.tar.gz (About 160 MB)
# Untar the file in the current directory

import numpy as np
import pickle
with open("batches.meta", 'rb') as fo:
    labelnames = pickle.load(fo)["label_names"]

data=[]
labels=[]
for i in range(1,6):
    with open("data_batch_{}".format(i), 'rb') as fo:
        alldata = pickle.load(fo, encoding='bytes')
    data.append(alldata[b"data"])
    labels.append(np.array(alldata[b"labels"]))
data=np.vstack(data)
labels=np.hstack(labels)

with open("test_batch", 'rb') as fo:
    alldata = pickle.load(fo, encoding='bytes')
testdata=alldata[b"data"]
testlabels=np.array(alldata[b"labels"])

#%% We now have two datasets, one for training and one for testing

print("Training data and labels with shapes: ",data.shape,labels.shape)
print("Testing data and labels with shapes: ",testdata.shape,testlabels.shape)
print("Meaning of the label codes: ",labelnames)

#%% Support function for drawing an image
import matplotlib.pyplot as plt
%matplotlib inline

def showim(im,l=None):
    """ Displays the image im (1D vector with 3072 elements)
    in the currently active axes, and set the title to the name
    corresponding to label l (if given)"""
    im=np.reshape(im,(3,32,32)) # reshape
    im=np.transpose(im, axes=(1,2,0)) # we want channels to be the last dimension, not the first.
    plt.imshow(im)
    if l is not None:
        plt.title(labelnames[l])
    
#%% Let's visualize some of our training images
for i in range(10):
    fig,ax = plt.subplots(figsize=(4,4),dpi=120)
    showim(data[i],labels[i])
    plt.show()
    
#%% Function to compute the euclidean distance between two images
def distance(im1,im2):
    """ Returns the euclidean distance between two vectors im1 and im2 """
    return np.linalg.norm(im1.astype(float).flatten()-im2.astype(float).flatten())

#%% Let's look for the closest (and farthest) training image to each of the first 20 testing images
    
for testi in range(20):
    im = testdata[testi]
    l = testlabels[testi]
    # im is now our test image, and l is the ground truth label
    
    d = np.full(len(data),np.nan)
    # for each training image, we'll save its distance from the query
    for i in range(len(data)):
        d[i] = distance(data[i],im)
    i_closest = np.argmin(d) # this is the index of the closest image
    i_farthest = np.argmax(d) # index of the farthest image
    
    print("Ground truth label: ",labelnames[l])
    print("Estimated label: ",labelnames[labels[i_closest]], flush=True)
    
    #%
    fig,(ax1,ax2,ax3) = plt.subplots(ncols=3,figsize=(14,6),dpi=150)
    plt.sca(ax1) # draw the test image (and label) on the first axis
    showim(testdata[testi],testlabels[testi])
    plt.sca(ax2) # draw the closest image (and label) on the second axis
    showim(data[i_closest],labels[i_closest])
    plt.sca(ax3) # draw the farthest image (and label) on the third axis
    showim(data[i_farthest],labels[i_farthest])
    
    plt.show()